@@ -44,8 +44,15 @@ Street, Fifth Floor, Boston, MA 02110-1335 USA
 #include <mysql.h>
 #include <mysqld.h>
 #include <my_sys.h>
+#include <stdlib.h>
 #include <string.h>
 #include <limits>
+#ifdef HAVE_PWD_H
+#ifdef HAVE_SYS_TYPES_H
+#include <sys/types.h>
+#endif
+#include <pwd.h>
+#endif
 #include "common.h"
 #include "xtrabackup.h"
 #include "srv0srv.h"
@@ -59,14 +66,13 @@ Street, Fifth Floor, Boston, MA 02110-1335 USA
 #include "page0zip.h"
 
 char *tool_name;
-char tool_args[2048];
+char tool_args[8192];
 
 /* mysql flavor and version */
 mysql_flavor_t server_flavor = FLAVOR_UNKNOWN;
 unsigned long mysql_server_version = 0;
 
 /* server capabilities */
-bool have_changed_page_bitmaps = false;
 bool have_backup_locks = false;
 bool have_lock_wait_timeout = false;
 bool have_galera_enabled = false;
@@ -75,7 +81,6 @@ bool have_multi_threaded_slave = false;
 bool have_gtid_slave = false;
 
 /* Kill long selects */
-os_thread_id_t	kill_query_thread_id;
 os_event_t	kill_query_thread_started;
 os_event_t	kill_query_thread_stopped;
 os_event_t	kill_query_thread_stop;
@@ -94,11 +99,54 @@ MYSQL *mysql_connection;
 
 extern my_bool opt_ssl_verify_server_cert, opt_use_ssl;
 
+
+/*
+  get_os_user()
+  Ressemles read_user_name() from libmariadb/libmariadb/mariadb_lib.c.
+*/
+
+#if !defined(_WIN32)
+
+#if defined(HAVE_GETPWUID) && defined(NO_GETPWUID_DECL)
+struct passwd *getpwuid(uid_t);
+char* getlogin(void);
+#endif
+
+static const char *get_os_user() // Posix
+{
+  if (!geteuid())
+    return "root";
+#ifdef HAVE_GETPWUID
+  struct passwd *pw;
+  const char *str;
+  if ((pw= getpwuid(geteuid())) != NULL)
+    return pw->pw_name;
+  if ((str= getlogin()) != NULL)
+    return str;
+#endif
+  if ((str= getenv("USER")) ||
+      (str= getenv("LOGNAME")) ||
+      (str= getenv("LOGIN")))
+    return str;
+  return NULL;
+}
+
+#else
+
+static const char *get_os_user() // Windows
+{
+  return getenv("USERNAME");
+}
+
+#endif // _WIN32
+
+
 MYSQL *
 xb_mysql_connect()
 {
 	MYSQL *connection = mysql_init(NULL);
 	char mysql_port_str[std::numeric_limits<int>::digits10 + 3];
+	const char *user= opt_user ? opt_user : get_os_user();
 
 	sprintf(mysql_port_str, "%d", opt_port);
 
@@ -108,6 +156,13 @@ xb_mysql_connect()
 		return(NULL);
 	}
 
+#if !defined(DONT_USE_MYSQL_PWD)
+	if (!opt_password)
+	{
+		opt_password=getenv("MYSQL_PWD");
+	}
+#endif
+
 	if (!opt_secure_auth) {
 		mysql_options(connection, MYSQL_SECURE_AUTH,
 			      (char *) &opt_secure_auth);
@@ -119,9 +174,9 @@ xb_mysql_connect()
 	mysql_options(connection, MYSQL_OPT_PROTOCOL, &opt_protocol);
 	mysql_options(connection,MYSQL_SET_CHARSET_NAME, "utf8");
 
-	msg("Connecting to MySQL server host: %s, user: %s, password: %s, "
+	msg("Connecting to server host: %s, user: %s, password: %s, "
 	       "port: %s, socket: %s", opt_host ? opt_host : "localhost",
-	       opt_user ? opt_user : "not set",
+	       user ? user : "not set",
 	       opt_password ? "set" : "not set",
 	       opt_port != 0 ? mysql_port_str : "not set",
 	       opt_socket ? opt_socket : "not set");
@@ -142,11 +197,11 @@ xb_mysql_connect()
 
 	if (!mysql_real_connect(connection,
 				opt_host ? opt_host : "localhost",
-				opt_user,
+				user,
 				opt_password,
 				"" /*database*/, opt_port,
 				opt_socket, 0)) {
-		msg("Failed to connect to MySQL server: %s.", mysql_error(connection));
+		msg("Failed to connect to server: %s.", mysql_error(connection));
 		mysql_close(connection);
 		return(NULL);
 	}
@@ -259,7 +314,8 @@ free_mysql_variables(mysql_variable *vars)
 
 static
 char *
-read_mysql_one_value(MYSQL *connection, const char *query)
+read_mysql_one_value(MYSQL *connection, const char *query,
+                     uint column, uint expect_columns)
 {
 	MYSQL_RES *mysql_result;
 	MYSQL_ROW row;
@@ -267,10 +323,10 @@ read_mysql_one_value(MYSQL *connection, const char *query)
 
 	mysql_result = xb_mysql_query(connection, query, true);
 
-	ut_ad(mysql_num_fields(mysql_result) == 1);
+	ut_ad(mysql_num_fields(mysql_result) == expect_columns);
 
 	if ((row = mysql_fetch_row(mysql_result))) {
-		result = strdup(row[0]);
+		result = strdup(row[column]);
 	}
 
 	mysql_free_result(mysql_result);
@@ -278,6 +334,15 @@ read_mysql_one_value(MYSQL *connection, const char *query)
 	return(result);
 }
 
+
+static
+char *
+read_mysql_one_value(MYSQL *mysql, const char *query)
+{
+  return read_mysql_one_value(mysql, query, 0/*offset*/, 1/*total columns*/);
+}
+
+
 static
 bool
 check_server_version(unsigned long version_number,
@@ -325,7 +390,7 @@ check_server_version(unsigned long version_number,
 }
 
 /*********************************************************************//**
-Receive options important for XtraBackup from MySQL server.
+Receive options important for XtraBackup from server.
 @return	true on success. */
 bool get_mysql_vars(MYSQL *connection)
 {
@@ -349,6 +414,7 @@ bool get_mysql_vars(MYSQL *connection)
   char *innodb_undo_directory_var= NULL;
   char *innodb_page_size_var= NULL;
   char *innodb_undo_tablespaces_var= NULL;
+  char *aria_log_dir_path_var= NULL;
   char *page_zip_level_var= NULL;
   char *ignore_db_dirs= NULL;
   char *endptr;
@@ -379,6 +445,7 @@ bool get_mysql_vars(MYSQL *connection)
       {"innodb_undo_tablespaces", &innodb_undo_tablespaces_var},
       {"innodb_compression_level", &page_zip_level_var},
       {"ignore_db_dirs", &ignore_db_dirs},
+      {"aria_log_dir_path", &aria_log_dir_path_var},
       {NULL, NULL}};
 
   read_mysql_variables(connection, "SHOW VARIABLES", mysql_vars, true);
@@ -510,6 +577,12 @@ bool get_mysql_vars(MYSQL *connection)
     ut_ad(*endptr == 0);
   }
 
+  if (aria_log_dir_path_var)
+  {
+    aria_log_dir_path= my_strdup(PSI_NOT_INSTRUMENTED,
+                                 aria_log_dir_path_var, MYF(MY_FAE));
+  }
+
   if (page_zip_level_var != NULL)
   {
     page_zip_level= static_cast<uint>(strtoul(page_zip_level_var, &endptr,
@@ -532,34 +605,6 @@ Query the server to find out what backup capabilities it supports.
 bool
 detect_mysql_capabilities_for_backup()
 {
-	const char *query = "SELECT 'INNODB_CHANGED_PAGES', COUNT(*) FROM "
-				"INFORMATION_SCHEMA.PLUGINS "
-			    "WHERE PLUGIN_NAME LIKE 'INNODB_CHANGED_PAGES'";
-	char *innodb_changed_pages = NULL;
-	mysql_variable vars[] = {
-		{"INNODB_CHANGED_PAGES", &innodb_changed_pages}, {NULL, NULL}};
-
-	if (xtrabackup_incremental) {
-
-		read_mysql_variables(mysql_connection, query, vars, true);
-
-		ut_ad(innodb_changed_pages != NULL);
-
-		have_changed_page_bitmaps = (atoi(innodb_changed_pages) == 1);
-
-		/* INNODB_CHANGED_PAGES are listed in
-		INFORMATION_SCHEMA.PLUGINS in MariaDB, but
-		FLUSH NO_WRITE_TO_BINLOG CHANGED_PAGE_BITMAPS
-		is not supported for versions below 10.1.6
-		(see MDEV-7472) */
-		if (server_flavor == FLAVOR_MARIADB &&
-		    mysql_server_version < 100106) {
-			have_changed_page_bitmaps = false;
-		}
-
-		free_mysql_variables(vars);
-	}
-
 	/* do some sanity checks */
 	if (opt_galera_info && !have_galera_enabled) {
 		msg("--galera-info is specified on the command "
@@ -856,7 +901,7 @@ start_query_killer()
 	kill_query_thread_started	= os_event_create(0);
 	kill_query_thread_stopped	= os_event_create(0);
 
-	os_thread_create(kill_query_thread, NULL, &kill_query_thread_id);
+	os_thread_create(kill_query_thread);
 
 	os_event_wait(kill_query_thread_started);
 }
@@ -913,12 +958,10 @@ bool lock_tables(MYSQL *connection)
 
   if (have_galera_enabled)
   {
-    xb_mysql_query(connection, "SET SESSION wsrep_causal_reads=0", false);
+    xb_mysql_query(connection, "SET SESSION wsrep_sync_wait=0", false);
   }
 
   xb_mysql_query(connection, "BACKUP STAGE START", true);
-  // xb_mysql_query(connection, "BACKUP STAGE FLUSH", true);
-  // xb_mysql_query(connection, "BACKUP STAGE BLOCK_DDL", true);
   xb_mysql_query(connection, "BACKUP STAGE BLOCK_COMMIT", true);
   /* Set the maximum supported session value for
   lock_wait_timeout to prevent unnecessary timeouts when the
@@ -1073,89 +1116,323 @@ cleanup:
 }
 
 
+class Var
+{
+  const char *m_name;
+  char *m_value;
+  /*
+    Disable copying constructors for safety, as the default binary copying
+    which would be wrong. If we ever want them, the m_value
+    member should be copied using an strdup()-alike function.
+  */
+  Var(const Var &); // Disabled
+  Var(Var &);       // Disabled
+public:
+  ~Var()
+  {
+    free(m_value);
+  }
+  Var(const char *name)
+   :m_name(name),
+    m_value(NULL)
+  { }
+  // Init using a SHOW VARIABLES LIKE 'name' query
+  Var(const char *name, MYSQL *mysql)
+   :m_name(name)
+  {
+    char buf[128];
+    my_snprintf(buf, sizeof(buf), "SHOW VARIABLES LIKE '%s'", m_name);
+    m_value= read_mysql_one_value(mysql, buf, 1/*offset*/, 2/*total columns*/);
+  }
+  /*
+    Init by name from a result set.
+    If the variable name is not found in the result set metadata field names,
+    it's value stays untouched.
+  */
+  bool init(MYSQL_RES *mysql_result, MYSQL_ROW row)
+  {
+    MYSQL_FIELD *field= mysql_fetch_fields(mysql_result);
+    for (uint i= 0; i < mysql_num_fields(mysql_result); i++)
+    {
+      if (!strcmp(field[i].name, m_name))
+      {
+        free(m_value); // In case it was initialized earlier
+        m_value= row[i] ? strdup(row[i]) : NULL;
+        return false;
+      }
+    }
+    return true;
+  }
+  void replace(char from, char to)
+  {
+    ut_ad(m_value);
+    for (char *ptr= strchr(m_value, from); ptr; ptr= strchr(ptr, from))
+      *ptr= to;
+  }
+
+  const char *value() const { return m_value; }
+  bool eq_value(const char *str, size_t length) const
+  {
+    return m_value && !strncmp(m_value, str, length) && m_value[length] == '\0';
+  }
+  bool is_null_or_empty() const { return !m_value || !m_value[0]; }
+  bool print(String *to) const
+  {
+    ut_ad(m_value);
+    return to->append(m_value);
+  }
+  bool print_quoted(String *to) const
+  {
+    ut_ad(m_value);
+    return to->append("'") || to->append(m_value) || to->append("'");
+  }
+  bool print_set_global(String *to) const
+  {
+    ut_ad(m_value);
+    return
+      to->append("SET GLOBAL ") ||
+      to->append(m_name) ||
+      to->append(" = '") ||
+      to->append(m_value) ||
+      to->append("';\n");
+  }
+};
+
+
+class Show_slave_status
+{
+  Var m_mariadb_connection_name; // MariaDB: e.g. 'master1'
+  Var m_master;              // e.g. 'localhost'
+  Var m_filename;            // e.g. 'source-bin.000002'
+  Var m_position;            // a number
+  Var m_mysql_gtid_executed; // MySQL56: e.g. single '<UUID>:1-5" or multiline
+                             // '<UUID1>:1-10,\n<UUID2>:1-20\n<UUID3>:1-30'
+  Var m_mariadb_using_gtid;  // MariaDB: 'No','Slave_Pos','Current_Pos'
+
+public:
+
+  Show_slave_status()
+   :m_mariadb_connection_name("Connection_name"),
+    m_master("Master_Host"),
+    m_filename("Relay_Master_Log_File"),
+    m_position("Exec_Master_Log_Pos"),
+    m_mysql_gtid_executed("Executed_Gtid_Set"),
+    m_mariadb_using_gtid("Using_Gtid")
+  { }
+
+  void init(MYSQL_RES *res, MYSQL_ROW row)
+  {
+    m_mariadb_connection_name.init(res, row);
+    m_master.init(res, row);
+    m_filename.init(res, row);
+    m_position.init(res, row);
+    m_mysql_gtid_executed.init(res, row);
+    m_mariadb_using_gtid.init(res, row);
+    // Normalize
+    if (m_mysql_gtid_executed.value())
+      m_mysql_gtid_executed.replace('\n', ' ');
+  }
+
+  static void msg_is_not_slave()
+  {
+    msg("Failed to get master binlog coordinates "
+        "from SHOW SLAVE STATUS.This means that the server is not a "
+        "replication slave. Ignoring the --slave-info option");
+  }
+
+  bool is_mariadb_using_gtid() const
+  {
+    return !m_mariadb_using_gtid.eq_value("No", 2);
+  }
+
+  static bool start_comment_chunk(String *to)
+  {
+    return to->length() ? to->append("; ") : false;
+  }
+
+  bool print_connection_name_if_set(String *to) const
+  {
+    if (!m_mariadb_connection_name.is_null_or_empty())
+      return m_mariadb_connection_name.print_quoted(to) || to->append(' ');
+    return false;
+  }
+
+  bool print_comment_master_identity(String *comment) const
+  {
+    if (comment->append("master "))
+      return true;
+    if (!m_mariadb_connection_name.is_null_or_empty())
+      return m_mariadb_connection_name.print_quoted(comment);
+    return comment->append("''"); // Default not named master
+  }
+
+  bool print_using_master_log_pos(String *sql, String *comment) const
+  {
+    return
+      sql->append("CHANGE MASTER ") ||
+      print_connection_name_if_set(sql) ||
+      sql->append("TO MASTER_LOG_FILE=") || m_filename.print_quoted(sql) ||
+      sql->append(", MASTER_LOG_POS=")   || m_position.print(sql) ||
+      sql->append(";\n") ||
+      print_comment_master_identity(comment) ||
+      comment->append(" filename ")  || m_filename.print_quoted(comment) ||
+      comment->append(" position ")  || m_position.print_quoted(comment);
+  }
+
+  bool print_mysql56(String *sql, String *comment) const
+  {
+    /*
+      SET @@GLOBAL.gtid_purged = '2174B383-5441-11E8-B90A-C80AA9429562:1-1029, '
+                                 '224DA167-0C0C-11E8-8442-00059A3C7B00:1-2695';
+      CHANGE MASTER TO MASTER_AUTO_POSITION=1;
+    */
+    return
+      sql->append("SET GLOBAL gtid_purged=") ||
+      m_mysql_gtid_executed.print_quoted(sql) ||
+      sql->append(";\n") ||
+      sql->append("CHANGE MASTER TO MASTER_AUTO_POSITION=1;\n") ||
+      print_comment_master_identity(comment) ||
+      comment->append(" purge list ") ||
+      m_mysql_gtid_executed.print_quoted(comment);
+  }
+
+  bool print_mariadb10_using_gtid(String *sql, String *comment) const
+  {
+    return
+      sql->append("CHANGE MASTER ") ||
+      print_connection_name_if_set(sql) ||
+      sql->append("TO master_use_gtid = slave_pos;\n") ||
+      print_comment_master_identity(comment) ||
+      comment->append(" master_use_gtid = slave_pos");
+  }
+
+  bool print(String *sql, String *comment, const Var &gtid_slave_pos) const
+  {
+    if (!m_mysql_gtid_executed.is_null_or_empty())
+    {
+      /* MySQL >= 5.6 with GTID enabled */
+      return print_mysql56(sql, comment);
+    }
+
+    if (!gtid_slave_pos.is_null_or_empty() && is_mariadb_using_gtid())
+    {
+      /* MariaDB >= 10.0 with GTID enabled */
+      return print_mariadb10_using_gtid(sql, comment);
+    }
+
+    return print_using_master_log_pos(sql, comment);
+  }
+
+  /*
+    Get master info into strings "sql" and "comment" from a MYSQL_RES.
+    @return false on success
+    @return true on error
+  */
+  static bool get_slave_info(MYSQL_RES *show_slave_info_result,
+                             const Var &gtid_slave_pos,
+                             String *sql, String *comment)
+  {
+    if (!gtid_slave_pos.is_null_or_empty())
+    {
+      // Print gtid_slave_pos if any of the masters really needs it.
+      while (MYSQL_ROW row= mysql_fetch_row(show_slave_info_result))
+      {
+        Show_slave_status status;
+        status.init(show_slave_info_result, row);
+        if (status.is_mariadb_using_gtid())
+        {
+          if (gtid_slave_pos.print_set_global(sql) ||
+              comment->append("gtid_slave_pos ") ||
+              gtid_slave_pos.print_quoted(comment))
+            return true; // Error
+          break;
+        }
+      }
+    }
+
+    // Print the list of masters
+    mysql_data_seek(show_slave_info_result, 0);
+    while (MYSQL_ROW row= mysql_fetch_row(show_slave_info_result))
+    {
+      Show_slave_status status;
+      status.init(show_slave_info_result, row);
+      if (start_comment_chunk(comment) ||
+          status.print(sql, comment, gtid_slave_pos))
+        return true; // Error
+    }
+    return false; // Success
+  }
+
+  /*
+    Get master info into strings "sql" and "comment".
+    @return false on success
+    @return true on error
+  */
+  static bool get_slave_info(MYSQL *mysql, bool show_all_slave_status,
+                             String *sql, String *comment)
+  {
+    bool rc= false; // Success
+    // gtid_slave_pos - MariaDB variable : e.g. "0-1-1" or "1-10-100,2-20-500"
+    Var gtid_slave_pos("gtid_slave_pos", mysql);
+    const char *query= show_all_slave_status ? "SHOW ALL SLAVES STATUS" :
+                                               "SHOW SLAVE STATUS";
+    MYSQL_RES *mysql_result= xb_mysql_query(mysql, query, true);
+    if (!mysql_num_rows(mysql_result))
+    {
+      msg_is_not_slave();
+      // Don't change rc, we still want to continue the backup
+    }
+    else
+    {
+      rc= get_slave_info(mysql_result, gtid_slave_pos, sql, comment);
+    }
+    mysql_free_result(mysql_result);
+    return rc;
+  }
+};
+
+
+
 /*********************************************************************//**
 Retrieves MySQL binlog position of the master server in a replication
 setup and saves it in a file. It also saves it in mysql_slave_position
-variable. */
+variable.
+@returns false on error
+@returns true on success
+*/
 bool
-write_slave_info(MYSQL *connection)
+write_slave_info(ds_ctxt *datasink, MYSQL *connection)
 {
-	char *master = NULL;
-	char *filename = NULL;
-	char *gtid_executed = NULL;
-	char *position = NULL;
-	char *gtid_slave_pos = NULL;
-	char *ptr;
-	bool result = false;
+  String sql, comment;
+  bool show_all_slaves_status= false;
 
-	mysql_variable status[] = {
-		{"Master_Host", &master},
-		{"Relay_Master_Log_File", &filename},
-		{"Exec_Master_Log_Pos", &position},
-		{"Executed_Gtid_Set", &gtid_executed},
-		{NULL, NULL}
-	};
+  switch (server_flavor)
+  {
+  case FLAVOR_MARIADB:
+    show_all_slaves_status= mysql_server_version >= 100000;
+    break;
+  case FLAVOR_UNKNOWN:
+  case FLAVOR_MYSQL:
+  case FLAVOR_PERCONA_SERVER:
+    break;
+  }
 
-	mysql_variable variables[] = {
-		{"gtid_slave_pos", &gtid_slave_pos},
-		{NULL, NULL}
-	};
+  if (Show_slave_status::get_slave_info(connection, show_all_slaves_status,
+                                        &sql, &comment))
+    return false; // Error
 
-	read_mysql_variables(connection, "SHOW SLAVE STATUS", status, false);
-	read_mysql_variables(connection, "SHOW VARIABLES", variables, true);
+  if (!sql.length())
+  {
+    /*
+      SHOW [ALL] SLAVE STATUS returned no rows.
+      Don't create the file, but return success to continue the backup.
+    */
+    return true; // Success
+  }
 
-	if (master == NULL || filename == NULL || position == NULL) {
-		msg("Failed to get master binlog coordinates "
-			"from SHOW SLAVE STATUS.This means that the server is not a "
-			"replication slave. Ignoring the --slave-info option");
-		/* we still want to continue the backup */
-		result = true;
-		goto cleanup;
-	}
-
-	/* Print slave status to a file.
-	If GTID mode is used, construct a CHANGE MASTER statement with
-	MASTER_AUTO_POSITION and correct a gtid_purged value. */
-	if (gtid_executed != NULL && *gtid_executed) {
-		/* MySQL >= 5.6 with GTID enabled */
-
-		for (ptr = strchr(gtid_executed, '\n');
-		     ptr;
-		     ptr = strchr(ptr, '\n')) {
-			*ptr = ' ';
-		}
-
-		result = backup_file_printf(XTRABACKUP_SLAVE_INFO,
-			"SET GLOBAL gtid_purged='%s';\n"
-			"CHANGE MASTER TO MASTER_AUTO_POSITION=1\n",
-			gtid_executed);
-
-		ut_a(asprintf(&mysql_slave_position,
-			"master host '%s', purge list '%s'",
-			master, gtid_executed) != -1);
-	} else if (gtid_slave_pos && *gtid_slave_pos) {
-		/* MariaDB >= 10.0 with GTID enabled */
-		result = backup_file_printf(XTRABACKUP_SLAVE_INFO,
-			"SET GLOBAL gtid_slave_pos = '%s';\n"
-			"CHANGE MASTER TO master_use_gtid = slave_pos\n",
-			gtid_slave_pos);
-		ut_a(asprintf(&mysql_slave_position,
-			"master host '%s', gtid_slave_pos %s",
-			master, gtid_slave_pos) != -1);
-	} else {
-		result = backup_file_printf(XTRABACKUP_SLAVE_INFO,
-			"CHANGE MASTER TO MASTER_LOG_FILE='%s', "
-			"MASTER_LOG_POS=%s\n", filename, position);
-		ut_a(asprintf(&mysql_slave_position,
-			"master host '%s', filename '%s', position '%s'",
-			master, filename, position) != -1);
-	}
-
-cleanup:
-	free_mysql_variables(status);
-	free_mysql_variables(variables);
-
-	return(result);
+  mysql_slave_position= strdup(comment.c_ptr());
+  return datasink->backup_file_print_buf(XTRABACKUP_SLAVE_INFO,
+                               sql.ptr(), sql.length());
 }
 
 
@@ -1163,10 +1440,11 @@ cleanup:
 Retrieves MySQL Galera and
 saves it in a file. It also prints it to stdout. */
 bool
-write_galera_info(MYSQL *connection)
+write_galera_info(ds_ctxt *datasink, MYSQL *connection)
 {
 	char *state_uuid = NULL, *state_uuid55 = NULL;
 	char *last_committed = NULL, *last_committed55 = NULL;
+	char *domain_id = NULL, *domain_id55 = NULL;
 	bool result;
 
 	mysql_variable status[] = {
@@ -1177,6 +1455,12 @@ write_galera_info(MYSQL *connection)
 		{NULL, NULL}
 	};
 
+	mysql_variable value[] = {
+		{"Wsrep_gtid_domain_id", &domain_id},
+		{"wsrep_gtid_domain_id", &domain_id55},
+		{NULL, NULL}
+	};
+
 	/* When backup locks are supported by the server, we should skip
 	creating xtrabackup_galera_info file on the backup stage, because
 	wsrep_local_state_uuid and wsrep_last_committed will be inconsistent
@@ -1190,14 +1474,35 @@ write_galera_info(MYSQL *connection)
 
 	if ((state_uuid == NULL && state_uuid55 == NULL)
 		|| (last_committed == NULL && last_committed55 == NULL)) {
-		msg("Failed to get master wsrep state from SHOW STATUS.");
-		result = false;
+		msg("Warning: failed to get master wsrep state from SHOW STATUS.");
+		result = true;
 		goto cleanup;
 	}
 
-	result = backup_file_printf(XTRABACKUP_GALERA_INFO,
-		"%s:%s\n", state_uuid ? state_uuid : state_uuid55,
-			last_committed ? last_committed : last_committed55);
+	read_mysql_variables(connection, "SHOW VARIABLES LIKE 'wsrep%'", value, true);
+
+	if (domain_id == NULL && domain_id55 == NULL) {
+		msg("Warning: failed to get master wsrep state from SHOW VARIABLES.");
+		result = true;
+		goto cleanup;
+	}
+
+	result = datasink->backup_file_printf(XTRABACKUP_GALERA_INFO,
+		"%s:%s %s\n", state_uuid ? state_uuid : state_uuid55,
+			      last_committed ? last_committed : last_committed55,
+			      domain_id ? domain_id : domain_id55);
+
+	if (result)
+	{
+	  result= datasink->backup_file_printf(XTRABACKUP_DONOR_GALERA_INFO,
+		"%s:%s %s\n", state_uuid ? state_uuid : state_uuid55,
+			      last_committed ? last_committed : last_committed55,
+			      domain_id ? domain_id : domain_id55);
+	}
+	if (result)
+	{
+		write_current_binlog_file(datasink, connection);
+	}
 
 cleanup:
 	free_mysql_variables(status);
@@ -1210,7 +1515,7 @@ cleanup:
 Flush and copy the current binary log file into the backup,
 if GTID is enabled */
 bool
-write_current_binlog_file(MYSQL *connection)
+write_current_binlog_file(ds_ctxt *datasink, MYSQL *connection)
 {
 	char *executed_gtid_set = NULL;
 	char *gtid_binlog_state = NULL;
@@ -1280,7 +1585,7 @@ write_current_binlog_file(MYSQL *connection)
 
 		snprintf(filepath, sizeof(filepath), "%s%c%s",
 			 log_bin_dir, FN_LIBCHAR, log_bin_file);
-		result = copy_file(ds_data, filepath, log_bin_file, 0);
+		result = datasink->copy_file(filepath, log_bin_file, 0);
 	}
 
 cleanup:
@@ -1296,7 +1601,7 @@ cleanup:
 Retrieves MySQL binlog position and
 saves it in a file. It also prints it to stdout. */
 bool
-write_binlog_info(MYSQL *connection)
+write_binlog_info(ds_ctxt *datasink, MYSQL *connection)
 {
 	char *filename = NULL;
 	char *position = NULL;
@@ -1341,14 +1646,14 @@ write_binlog_info(MYSQL *connection)
 			"filename '%s', position '%s', "
 			"GTID of the last change '%s'",
 			filename, position, gtid) != -1);
-		result = backup_file_printf(XTRABACKUP_BINLOG_INFO,
+		result = datasink->backup_file_printf(XTRABACKUP_BINLOG_INFO,
 					    "%s\t%s\t%s\n", filename, position,
 					    gtid);
 	} else {
 		ut_a(asprintf(&mysql_binlog_position,
 			"filename '%s', position '%s'",
 			filename, position) != -1);
-		result = backup_file_printf(XTRABACKUP_BINLOG_INFO,
+		result = datasink->backup_file_printf(XTRABACKUP_BINLOG_INFO,
 					    "%s\t%s\n", filename, position);
 	}
 
@@ -1388,8 +1693,9 @@ PERCONA_SCHEMA.xtrabackup_history and writes a new history record to the
 table containing all the history info particular to the just completed
 backup. */
 bool
-write_xtrabackup_info(MYSQL *connection, const char * filename, bool history,
-                       bool stream)
+write_xtrabackup_info(ds_ctxt *datasink,
+                      MYSQL *connection, const char * filename, bool history,
+                      bool stream)
 {
 
 	bool result = true;
@@ -1465,7 +1771,7 @@ write_xtrabackup_info(MYSQL *connection, const char * filename, bool history,
 	}
 
 	if (stream) {
-		backup_file_printf(filename, "%s", buf);
+		datasink->backup_file_printf(filename, "%s", buf);
 	} else {
 		fp = fopen(filename, "w");
 		if (!fp) {
@@ -1586,11 +1892,11 @@ static std::string make_local_paths(const char *data_file_path)
 	return buf.str();
 }
 
-bool write_backup_config_file()
+bool write_backup_config_file(ds_ctxt *datasink)
 {
-	int rc= backup_file_printf("backup-my.cnf",
-		"# This MySQL options file was generated by innobackupex.\n\n"
-		"# The MySQL server\n"
+	int rc= datasink->backup_file_printf("backup-my.cnf",
+		"# This options file was generated by innobackupex.\n\n"
+		"# The server\n"
 		"[mysqld]\n"
 		"innodb_checksum_algorithm=%s\n"
 		"innodb_data_file_path=%s\n"
@@ -1631,9 +1937,11 @@ char *make_argv(char *buf, size_t len, int argc, char **argv)
 		if (strncmp(*argv, "--password", strlen("--password")) == 0) {
 			arg = "--password=...";
 		}
-		left-= snprintf(buf + len - left, left,
+		uint l= snprintf(buf + len - left, left,
 				"%s%c", arg, argc > 1 ? ' ' : 0);
 		++argv; --argc;
+                if (l < left)
+                  left-= l;
 	}
 
 	return buf;
@@ -1662,20 +1970,8 @@ select_history()
 	return(true);
 }
 
-bool
-flush_changed_page_bitmaps()
-{
-	if (xtrabackup_incremental && have_changed_page_bitmaps &&
-	    !xtrabackup_incremental_force_scan) {
-		xb_mysql_query(mysql_connection,
-			"FLUSH NO_WRITE_TO_BINLOG CHANGED_PAGE_BITMAPS", false);
-	}
-	return(true);
-}
-
-
 /*********************************************************************//**
-Deallocate memory, disconnect from MySQL server, etc.
+Deallocate memory, disconnect from server, etc.
 @return	true on success. */
 void
 backup_cleanup()
