@@ -229,7 +229,7 @@ bool Inet6::ascii_to_ipv6(const char *str, size_t str_length)
         continue;
       }
 
-      if (!*p || p >= str_end)
+      if (p >= str_end || !*p)
       {
         DBUG_PRINT("error", ("ascii_to_ipv6(%.*s): invalid IPv6 address: "
                              "ending at ':'.", (int) str_length, str));
@@ -936,8 +936,11 @@ public:
       Mixing of two different non-traditional types is currently prevented.
       This may change in the future. For example, INET4 and INET6
       data types can be made comparable.
+      But we allow mixing INET6 to a data type directly inherited from
+      a traditional type, e.g. INET6=VARCHAR/JSON.
     */
-    DBUG_ASSERT(item->type_handler()->is_traditional_scalar_type() ||
+    DBUG_ASSERT(item->type_handler()->type_handler_base_or_self()->
+                  is_traditional_scalar_type() ||
                 item->type_handler() == type_handler());
     return true;
   }
@@ -951,7 +954,8 @@ public:
                           bool is_eq_func) const override
   {
     // See the DBUG_ASSERT comment in can_optimize_keypart_ref()
-    DBUG_ASSERT(item->type_handler()->is_traditional_scalar_type() ||
+    DBUG_ASSERT(item->type_handler()->type_handler_base_or_self()->
+                  is_traditional_scalar_type() ||
                 item->type_handler() == type_handler());
     return true;
   }
@@ -1105,9 +1109,14 @@ public:
     if (!example)
       return false;
     value_cached= true;
-    null_value= example->val_native_with_conversion_result(current_thd,
-                                                           &m_value,
-                                                           type_handler());
+    /*
+      Merge comments: in 10.7 this code migrated to
+      Item_cache_fbt in to sql/sql_type_fixedbin.h
+    */
+    null_value_inside= null_value=
+      example->val_native_with_conversion_result(current_thd,
+                                                 &m_value,
+                                                 type_handler());
     return true;
   }
   String* val_str(String *to)
@@ -1294,14 +1303,22 @@ public:
   {
     return &type_handler_inet6;
   }
-  void set(uint pos, Item *item) override
+  bool set(uint pos, Item *item) override
   {
+    /*
+      A 10.5->10.6 merge hint:
+      This method migrated to sql/sql_type_fixedbin.h,
+      to the class in_fbt.
+    */
     Inet6 *buff= &((Inet6 *) base)[pos];
     Inet6_null value(item);
     if (value.is_null())
+    {
       *buff= Inet6_zero();
-    else
-      *buff= value;
+      return true;
+    }
+    *buff= value;
+    return false;
   }
   uchar *get_value(Item *item) override
   {
@@ -1434,7 +1451,7 @@ void Type_handler_inet6::Item_param_setup_conversion(THD *thd,
 
 void Type_handler_inet6::make_sort_key_part(uchar *to, Item *item,
                                             const SORT_FIELD_ATTR *sort_field,
-                                            Sort_param *param) const
+                                            String *tmp_str) const
 {
   DBUG_ASSERT(item->type_handler() == this);
   NativeBufferInet6 tmp;
@@ -1457,7 +1474,7 @@ void Type_handler_inet6::make_sort_key_part(uchar *to, Item *item,
 uint
 Type_handler_inet6::make_packed_sort_key_part(uchar *to, Item *item,
                                             const SORT_FIELD_ATTR *sort_field,
-                                            Sort_param *param) const
+                                            String *tmp_str) const
 {
   DBUG_ASSERT(item->type_handler() == this);
   NativeBufferInet6 tmp;
