@@ -1,6 +1,6 @@
 /*
    Copyright (c) 2000, 2019, Oracle and/or its affiliates.
-   Copyright (c) 2010, 2021, MariaDB
+   Copyright (c) 2010, 2022, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -254,7 +254,10 @@ int update_portion_of_time(THD *thd, TABLE *table,
     res= src->save_in_field(table->field[dst_fieldno], true);
 
   if (likely(!res))
+  {
+    table->period_prepare_autoinc();
     res= table->update_generated_fields();
+  }
 
   if(likely(!res))
     res= table->file->ha_update_row(table->record[1], table->record[0]);
@@ -406,7 +409,8 @@ bool mysql_delete(THD *thd, TABLE_LIST *table_list, COND *conds,
     DBUG_RETURN(TRUE);
 
   const_cond= (!conds || conds->const_item());
-  safe_update= MY_TEST(thd->variables.option_bits & OPTION_SAFE_UPDATES);
+  safe_update= (thd->variables.option_bits & OPTION_SAFE_UPDATES) &&
+               !thd->lex->describe;
   if (safe_update && const_cond)
   {
     my_message(ER_UPDATE_WITHOUT_KEY_IN_SAFE_MODE,
@@ -531,7 +535,7 @@ bool mysql_delete(THD *thd, TABLE_LIST *table_list, COND *conds,
   }
 
   /* If running in safe sql mode, don't allow updates without keys */
-  if (table->opt_range_keys.is_clear_all())
+  if (!select || !select->quick)
   {
     thd->set_status_no_index_used();
     if (safe_update && !using_limit)
@@ -692,6 +696,10 @@ bool mysql_delete(THD *thd, TABLE_LIST *table_list, COND *conds,
   {
     table->use_all_columns();
     table->rpl_write_set= table->write_set;
+    // Initialize autoinc.
+    // We don't set next_number_field here, as it is handled manually.
+    if (table->found_next_number_field)
+      table->file->info(HA_STATUS_AUTO);
   }
   else
   {
@@ -1065,6 +1073,8 @@ int mysql_prepare_delete(THD *thd, TABLE_LIST *table_list, Item **conds,
     DBUG_RETURN(TRUE);
 
   select_lex->fix_prepare_information(thd, conds, &fake_conds);
+  if (!thd->lex->upd_del_where)
+    thd->lex->upd_del_where= *conds;
   DBUG_RETURN(FALSE);
 }
 
@@ -1227,6 +1237,13 @@ multi_delete::initialize_tables(JOIN *join)
   {
     TABLE_LIST *tbl= walk->correspondent_table->find_table_for_update();
     tables_to_delete_from|= tbl->table->map;
+
+    /*
+      Ensure that filesort re-reads the row from the engine before
+      delete is called.
+    */
+    join->map2table[tbl->table->tablenr]->keep_current_rowid= true;
+
     if (delete_while_scanning &&
         unique_table(thd, tbl, join->tables_list, 0))
     {
@@ -1291,7 +1308,9 @@ multi_delete::initialize_tables(JOIN *join)
                                                   table->file->ref_length,
                                                   MEM_STRIP_BUF_SIZE);
   }
-  init_ftfuncs(thd, thd->lex->current_select, 1);
+  if (init_ftfuncs(thd, thd->lex->current_select, 1))
+    DBUG_RETURN(true);
+
   DBUG_RETURN(thd->is_fatal_error);
 }
 
