@@ -47,6 +47,58 @@ namespace tpool
   static const std::chrono::milliseconds LONG_TASK_DURATION = std::chrono::milliseconds(500);
   static const int  OVERSUBSCRIBE_FACTOR = 2;
 
+/**
+  Process the cb synchronously
+*/
+void aio::synchronous(aiocb *cb)
+{
+#ifdef _WIN32
+  size_t ret_len;
+#else
+  ssize_t ret_len;
+#endif
+  int err= 0;
+  switch (cb->m_opcode)
+  {
+  case aio_opcode::AIO_PREAD:
+    ret_len= pread(cb->m_fh, cb->m_buffer, cb->m_len, cb->m_offset);
+    break;
+  case aio_opcode::AIO_PWRITE:
+    ret_len= pwrite(cb->m_fh, cb->m_buffer, cb->m_len, cb->m_offset);
+    break;
+  default:
+    abort();
+  }
+#ifdef _WIN32
+  if (static_cast<int>(ret_len) < 0)
+    err= GetLastError();
+#else
+  if (ret_len < 0)
+  {
+    err= errno;
+    ret_len= 0;
+  }
+#endif
+  cb->m_ret_len = ret_len;
+  cb->m_err = err;
+  if (!err && cb->m_ret_len != cb->m_len)
+    finish_synchronous(cb);
+}
+
+
+/**
+  A partial read/write has occured, continue synchronously.
+*/
+void aio::finish_synchronous(aiocb *cb)
+{
+  assert(cb->m_ret_len != (unsigned int) cb->m_len && !cb->m_err);
+  /* partial read/write */
+  cb->m_buffer= (char *) cb->m_buffer + cb->m_ret_len;
+  cb->m_len-= (unsigned int) cb->m_ret_len;
+  cb->m_offset+= cb->m_ret_len;
+  synchronous(cb);
+}
+
 /**
   Implementation of generic threadpool.
   This threadpool consists of the following components
@@ -293,22 +345,23 @@ public:
     int m_period;
     std::mutex m_mtx;
     bool m_on;
-    std::atomic<bool> m_running;
+    std::atomic<int> m_running;
 
     void run()
     {
       /*
         In rare cases, multiple callbacks can be scheduled,
-        e.g with set_time(0,0) in a loop.
-        We do not allow parallel execution, as user is not prepared.
+        at the same time,. e.g with set_time(0,0) in a loop.
+        We do not allow parallel execution, since it is against the expectations.
       */
-      bool expected = false;
-      if (!m_running.compare_exchange_strong(expected, true))
+      if (m_running.fetch_add(1, std::memory_order_acquire) > 0)
         return;
-
-      m_callback(m_data);
-      dbug_execute_after_task_callback();
-      m_running = false;
+      do
+      {
+        m_callback(m_data);
+        dbug_execute_after_task_callback();
+      }
+      while (m_running.fetch_sub(1, std::memory_order_release) != 1);
 
       if (m_pool && m_period)
       {
@@ -591,7 +644,7 @@ void thread_pool_generic::check_idle(std::chrono::system_clock::time_point now)
   }
 
   /* Switch timer off after 1 minute of idle time */
-  if (now - idle_since > max_idle_time)
+  if (now - idle_since > max_idle_time && m_active_threads.empty())
   {
     idle_since= invalid_timestamp;
     switch_timer(timer_state_t::OFF);
@@ -690,6 +743,12 @@ bool thread_pool_generic::add_thread()
   if (n_threads >= m_max_threads)
     return false;
 
+  /*
+    Deadlock danger exists, so monitor pool health
+    with maintenance timer.
+  */
+  switch_timer(timer_state_t::ON);
+
   if (n_threads >= m_min_threads)
   {
     auto now = std::chrono::system_clock::now();
@@ -700,8 +759,6 @@ bool thread_pool_generic::add_thread()
         Throttle thread creation and wakeup deadlock detection timer,
         if is it off.
       */
-      switch_timer(timer_state_t::ON);
-
       return false;
     }
   }
@@ -784,6 +841,7 @@ thread_pool_generic::thread_pool_generic(int min_threads, int max_threads) :
 
   // start the timer
   m_maintenance_timer.set_time(0, (int)m_timer_interval.count());
+  m_timer_state = timer_state_t::ON;
 }
 
 
